#ifndef __SATELLITE_H
#define __SATELLITE_H

// C++ common head file
#if defined(_MSC_VER) && _MSC_VER < 1300
	#include <iostream.h>
	#include <fstream.h>
	#include <math.h>
	#include <float.h>
	#include <list>
	#include <queue>
#else
	#include <fstream>
	#include <iostream>
	using namespace std;
	#include <cmath>
	#include <limits>
	#include <list>
	#include <queue>
	#include <stdexcept>
#endif

#include "constant.h"
using namespace Constant;
#include "config.h"
#include "CEngine.h"
#include "CentralBody.h"
#include "DE405.h"
#include "Coordination.h"
#include "StopCondition.h"
#include "Maneuver.h"
#include "Quaternion.h"
#include "Kepler.h"

// 㶯ö

// 
#define ODP_EARTH_ZONAL             0x01  // г
#define ODP_EARTH_TESSERAL          0x02  // г
#define ODP_AIR_DRAG                0x04  // 
#define ODP_LUNAR_CENT              0x08  // (Է)
#define ODP_SOLAR_CENT              0x10  // ̫(ԷĹ)
#define ODP_SOLAR_PRESSURE          0x20  // ̫ѹ
#define ODP_EARTH_ALBEDO            0x40  // չѹ
#define ODP_SOLID_TIDES             0x80  // 峱
#define ODP_OCEAN_TIDES             0x0100 // 򺣳
#define ODP_POSTNEWTON              0x0200  // ţЧӦ
#define ODP_ADDITIONAL              0x0400  // ϵ㶯
#define ODP_THERMAL_RADIATION_PRESSURE 0x0800 // ȷѹ
// 
#define ODP_LUNAR_ZONAL        0x1000 // г
#define ODP_LUNAR_TESSERAL     0x2000 // г
#define ODP_LUNAR_NON_SPHERE   0x3000 // (г+г)
#define ODP_EARTH_CENT         0x4000 // (Էǵ)
#define ODP_LUNAR_TIDES        0x8000 // 峱
#define ODP_LUNAR_LIBRATION    0x010000 // ƽ
// ̽
#define ODP_VENUS_CENT         0x100000 // 
#define ODP_MARS_CENT          0x200000 // 
#define ODP_JUPITER_CENT       0x400000 // ľ
#define ODP_SATURN_CENT        0x800000 // 
// ü
#define ODP_EARTH_ALL  0x0FFF // ֮㶯
#define ODP_GEO        0x0FFB // ͬͨǵ㶯
#define ODP_LEO        0x0007 // ͹ͨǵ㶯
#define ODP_LUNAR_SAT  0x1F030 // 

/*!̬ģʽ
\enum AttitudeMode
EarthPoint: Եض̬\n
SunPoint: ն̬\n
InertiallyFixed: Զ̬\n
Custom: Լ
*/
enum AttitudeMode{ EarthPoint = 0, SunPoint , InertiallyFixed , Custom };

/*! ǵ
\author 
CSatelliteBaseǵ

ǵĻ
- ʱ(ԷʼԪStartEpoch)ʱ()
- ϵλٶ,ĹϵΪJ2000.0,ĹϵΪƽΪ,xΪJ2000.0ϵͶӰ
- ƣǺʹ洢ļ
- ǵĹ
- Ƕѧԣʱȡ䡢

ʵֵĺҪΪȡֵRKF78㷨⺯ΪǶѧ̺͹ļ

㶯ԴѡʹSetForce(int l,unsigned char PTB)ʵ֣lΪPTBΪ㶯Դѡº궨壺
\code
#define ODP_ZONAL      0x80   // г
#define ODP_TESSERAL   0xC0   // г+г
#define ODP_LUNAR      0x20   // 
#define ODP_SOLAR      0x10   // ̫
#define ODP_LIGHT      0x08   // ̫ѹ
#define ODP_AIR        0x04   // 
#define ODP_POSTNEWTON 0x02   // ţЧӦ
#define ODP_LUNAR_NON_SPHERE  0x01  // 

#define ODP_ALL        0xFE   // ֮㶯
#define ODP_GEO        0xF4   // ͬͨǵ㶯гѹţЧӦ
#define ODP_LEO        0x08   // ͹ͨǵ㶯
\endcode
PTBΪһ8bit־֣8λֱʾ\n
||1||2||3||4||5||6||7||8
|Zonalг|Zonal+Tesseralгг|Lunar|Solar̫|Light̫ѹ|Air|PostNţЧӦ|Lunar_nonsphere
*/
class ORBITDYN_API CSatelliteBase
{
protected:
	//! ʼԪ
	CDateTime Epoch; //TODO: ӦTDTʱ䣬ҪUTC!!
	//! ʼԪʱ
	double ElapsedTime;
	//!ϵλ
	Vector Position;
	//!ϵٶ
	Vector Velocity;
	//!
	PLANET center;
public:
	//!ƣݣĬΪabc......
	std::string Name;
	//!ʼ
	Kepler Status0;
	//!ʼ(λ:kg)(Ĭ1000kg)
	double Mass0;
	//! ȼ(λ:kg)(ֵ0kg)
	double BurnedFuel;
	//!ϵ
	double Cd;
	//!ӭ
	double AirDragArea;
	//!̫ѹ
	double Cr;
	//!ѹЧ
	double LightPressArea;

	//!״̬־(true:,false:ػ)
	bool ThrustIsOn;
	//! ڱϵķ(һΪλʸ)
	Vector ThrustDirection;
	//! ̬Ԫ
	CQuaternion qbi;
	//! ̬ٶ
	Vector wbi;

	// ״̬:
	//!˲ʱİ볤(unitkm)
	double a;
	//!˲ʱƫ
	double e;
	//!˲ʱĹ(unit:rad)
	double i;
	//!˲ʱྭ(unit:rad)
	double Omega;
	//!˲ʱص(unit:rad)
	double w;
	//!˲ʱƽ(unit:rad)
	double M;

	//!˲ʱƽ(unit:rad)
	double f;
	//!˲ʱƽƫ(unit:rad)
	double E;
	//!˲ʱǷu=f+w     (unit:rad)
	double u;
	//!ƽٶn=sqrt(GE/a/a/a)   (unit:rad/s)
	double n;
	//! ʸ
	double r;
	//! ٶ
	double v;
	//! ٶ
	double w0;
	//! ϵϵת
	Matrix Coi;
	//! ٶ
	Vector acc;

///////  //////////////////////////

	//! ǳʼԪ
	inline CDateTime StartEpoch() const { return Epoch; }
	
	//! ǵǰԪ
	inline CDateTime CurrentEpoch() const { return Epoch+ElapsedTime; }

	//! ǵʱ䣬Epochʼ
	inline double t() const { return ElapsedTime; }

	//! ǵλ(λ:km)
	inline Vector Pos() const { return Position; }

	//! ǵٶ(λ:km/s)
	inline Vector Vel() const { return Velocity; }
	
	//! λ(λ:km)ٶ(λ:km/s)
	inline void SetPosVel(double t,const Vector& p,const Vector& v) 
	{
		ElapsedTime = t;
		Position = p;
		Velocity = v;
		RefreshStatus();
	}	

	//! ǵĹһKepler
	inline Kepler GetOrbitElements()const {	return Kepler(a,e,i,Omega,w,M);	}
	
	//! ǵĹһKepler
	inline void GetOrbitElements(double oe[6])const { memcpy(&oe,&a,sizeof(double)*6); }
	
	//! 
	inline double Mass(){ return Mass0 - BurnedFuel; }
	
	//! Ƿ
	inline double Force(){ return Engine->F; }

	CSatelliteBase();
	CSatelliteBase(std::string name);
	virtual ~CSatelliteBase();
	void Initialize();
	void Initialize(const CDateTime& t,const Vector& r,const Vector& v);
	void Initialize( const char* filename );
	void SetAutoSave(bool save = true);
	void SetForce(int l,unsigned int PTB);
	void SetEngine(double I,double f);

// ƺ
	void Propagate(double Step,const double Duration);
	void Propagate2Perigee();
	void Propagate2Apogee();
	void Propagate2AscendingNode();
	void Propagate2DescendingNode();
	void Propagate2Epoch(const CDateTime epoch);
	// 
	void PropagateBackward(double h,const double dt);

	//! ˲ʱ
	void ImpluseManeuver(Vector Dv,Coordination d = VVLH );

	//! رԶļ
	void CloseFile(){FileInstElem.close();FirstTimeOutput = true;}
private:
	//! CSatelliteBaseļԶļ
	static unsigned int Counter;

protected:
	
	//! ггĽ
	int LL;
	
	//! 㶯Դѡ־
	unsigned int Perturbation;
	
	//! 
	CEngine* Engine;

/////// ļ  ///////////////

	//! ļ
	fstream FileInstElem;
	
	//! Ƿһдļ־
	bool FirstTimeOutput;
	
	//! ǷԶ־
	bool Save;
	
	virtual void SaveElem() = 0;

	virtual int OneStep(double StepSize,double MaxCutError = 1,CStopCondition* SC = NULL);
	virtual void RefreshStatus() = 0;
	virtual void DynFunction(const double t,const double* x,double* y,int dim) const = 0;

	ORBITDYN_API friend double RKF78(const CSatelliteBase& s,double hh,double* x,double* xh,const int dim);
	ORBITDYN_API friend double Gill4(const CSatelliteBase& s,double hh,double* x,double* xh,const int dim);
};

ORBITDYN_API ostream & operator<<(ostream & os, const CSatelliteBase& sat);

/*!
\class CSatellite
CSatelliteBaseĻʵ
һǵĹ㣬ҲԼתƹͳӶ
\example examples/basic.cpp
*/
class ORBITDYN_API CSatellite:public CSatelliteBase
{
public:
	//! 켰
	CSatellite(){}
	~CSatellite(){}

	//! ʼú
	void Initialize(const CDateTime& t,const Kepler elem,const char flag = 'i');
	
	//! ǵĳʼֵļ
	void ReportInitial();

	//! عϵECFλ
	const Vector ECFPos();

	//! ڵع̵λٶ
	void GetECF(Vector& ECFr,Vector& ECFv);
	
	//! ǵĵγȺµĵγ
	CSpherical SubSatPoint(CSpherical & Geodetic,CSpherical & SSP);
	
	//! ƽ
	Kepler MedianElement() const;
	
	//! 
	CSpherical GetLLA();
	
	//double FiniteManeuver(Direction , StopCondition);

private:
	void DynFunction(const double t,const double* x,double* y,int dim) const;
	void RefreshStatus();

protected:
	void SaveElem();
};

/*!
ӿٶȵĹ
ʹGill4㷨̵ֹĶѧв¶λá㶯Ȳ
ÿһʱ

Ϊ˱֤ʱʱۼƵľȣʹĺElapsedTimemsΪʱԭеElapsedTimeΪ
*/
class ORBITDYN_API CRapidSatellite : public CSatellite
{
public:
	CRapidSatellite():ElapsedTimems(0)
		,m_lasttime(0)
		,m_lastpert(0)
		,PeriodCoord(1800000)
		,PeriodPert(50)
	{}
	~CRapidSatellite(){}

	//! ʼú
	void Initialize(const CDateTime& t,const Kepler elem,const char flag);

	// hλΪms
	void Propagate(int h);

private:
	// StepSizeλΪms
	int OneStep(int StepSize,double MaxCutError = 1,CStopCondition* StopCon = NULL );
	void DynFunction(const double t,const double* x,double* y,int dim) const;
	void RefreshStatus();
	void RefreshMatrix(const CDateTime t); // ˢ¶Ⱦ
	void RefreshPert(const CDateTime t);// ˢ㶯

	//! ԳʼԪʱ
	unsigned long long ElapsedTimems;

public:

	// Ϊ˼ӿٶȣ±RKFöѧʱ±ɲظ
	// 
	long PeriodCoord;   // 㶯ĸڣĬֵ1,800,000ms

private:
	//! ϴβʱ,±ÿPeriodCoord(ms)һ
	unsigned long long m_lasttime; 
	//! 
	Matrix PR;
	//! ¶
	Matrix NU;
	//! ƾ
	Matrix PW;
	//! J2000-->ECIת
	Matrix HG;
	//! λ
	Vector rm;
	//! ̫λ
	Vector rs;

public:
	// ¼ǰ㶯ٶȣRKFÿεöѧʱΪҶ㶯ٶȲ
	long PeriodPert;   // 㶯ĸڣĬֵ50ms

private:
	unsigned long long  m_lastpert; // ϴμ㶯ʱ䣬㶯ÿPeriodPert(ms)һ
	Vector F_zonal;     // г
	Vector F_tesseral;  // г
	Vector F_moon;      // 
	Vector F_sun;       // ̫
	Vector F_air;       // 
	Vector F_light;     // ѹ
	Vector F_ppn;       // ţ
	Vector F_Thrust;    // 
};

/*!
\class CLEO
͹
ʹСƫʹΪ״̬л
һ㲻ʹãΪ˱Ļ֣ΪԺоȱȽϱ
*/
class ORBITDYN_API CLEO:public CSatellite
{
public:
	CLEO(){}
	~CLEO(){}
	void Propagate(double h,const double dt);
private:
	int OneStep(double StepSize,double MaxCutError = 1.0,CStopCondition* SC = NULL);
	void DynFunction(const double t,const double* x,double* y,int dim) const;
	void RefreshStatus();
};
/*!
\class CGEO
ͬ
ӵͬõĹûоʵ
*/
class ORBITDYN_API CGEO:public CSatellite
{
public:
	double D;      //! Ư
	double ex,ey;  //! ƫʸ
	double ix,iy;  //! ʸ
	double L;      //! L=Omega+w+M
};
/*!
\class CMoonSat
ǵĹ
·еǵĹ

*/
class ORBITDYN_API CMoonSat: public CSatelliteBase
{
private:
	// Զѧ̺͹ļ¶
	void DynFunction(const double t,const double* x,double* y,int dim) const;
	void RefreshStatus();
	void SaveElem();
public:
	CMoonSat(){
		// ODP_LUNARʾǷǵ
		Perturbation = ODP_LUNAR_SAT;
		LL = 70;
		LightPressArea = 15.076;
		Mass0 = 1460.0;
		Cr = 1.44;
	}
	//! ʹĹϵʼ
	void Initialize(const CDateTime& t,const Kepler elem);
	//! ʹĹϵλٶȳʼ
	void Initialize(const CDateTime& t,const Vector p,const Vector v);
};

class ORBITDYN_API SatelliteException : public BaseException
{
public:
	SatelliteException(const std::string& details = "") 
		: BaseException("CSatellite exception:",details)
	{
	}
	virtual ~SatelliteException()
	{
	}
	SatelliteException(const SatelliteException& cdte)
		:BaseException(cdte)
	{
	}
};

#endif  // __SATELLITE_H
