//////////////////////////////////////////////////////////////
//File :  TVector.h
//Author: HanDle
//Date:   2004-8-31
//Usage:  Vector Calculation
//////////////////////////////////////////////////////////////

#ifndef __VECTOR_H
#define __VECTOR_H

#include "config.h"
#include "BaseException.h"

#if defined _MSC_VER && _MSC_VER < 1300
	#include <iostream.h>
	#include <math.h>
	#include <stdlib.h>
#else
	#include <iostream>
	#include <cmath>
#endif // _MSC_VER

#include "matrixTCL.h"

#define VEC_TEMPLATE

class ORBITDYN_API TVectorException : public BaseException//
{
public:
	TVectorException(const std::string& details) 
		:BaseException("TVector Exception Thrown: ",details)
	{
	}
	virtual ~TVectorException()
	{
	}
	TVectorException(const TVectorException& tve)
		:BaseException(tve)
	{
	}
};
#define REPORT_TVECTORERROR(a) throw TVectorException(a);

//ඨ
VEC_TEMPLATE
class ORBITDYN_API Vector
{
public:
	double x,y,z;//

	Vector(double d1=0, double d2=0, double d3=0):x(d1),y(d2),z(d3){ };      // Ĭϡ
	Vector(const Matrix& m){ x=m(0,0); y=m(1,0); z=m(2,0); }  // þ
	~Vector(){};//

	Vector& operator=(const Vector& v); // ֵ
	//Vector& operator=(const matrix& m); // ֵ
	operator Matrix () const;                 // תΪ31

// Subscript operator
	double operator()(size_t a) const;
	double& operator()(size_t a);
	double operator[](size_t a) const;
	double& operator[](size_t a);

	void SetValue(double n1, double n2, double n3);     // ֵ
	void SetValue(double * ptr);              // ʼ
	//inline Vector I(){ x=1; y=0; z=0; return *this; }
	//inline Vector J(){ x=0; y=1; z=0; return *this; }
	//inline Vector K(){ x=0; y=0; z=1; return *this; }
	double Module() const;                     //ģ
	Vector RotationX(double angle) const;  // xת
	Vector RotationY(double angle) const;  // yת
	Vector RotationZ(double angle) const;  // zת
	double* ToArray(double* ptr) const;                  //
	void Unit(); //λ
	void Zero(); // 
	const Matrix CrossMatrix();
};

namespace Constant
{
	const Vector Vector_i(1,0,0);
	const Vector Vector_j(0,1,0);
	const Vector Vector_k(0,0,1);
}

VEC_TEMPLATE ORBITDYN_API double Dot(const Vector& v1,const Vector& v2);//
VEC_TEMPLATE ORBITDYN_API double VVAngle(const Vector& v1,const Vector& v2);//ʸн
VEC_TEMPLATE ORBITDYN_API Vector Cross(const Vector &v1,const Vector& v2);//


//////////////////////////////////////////////////////////////////////////
//**********************
//ǳԱ
//**********************
VEC_TEMPLATE ORBITDYN_API Vector operator+(const Vector& v1,const Vector& v2); //ʸ
VEC_TEMPLATE ORBITDYN_API Vector operator+=(Vector & v1,const Vector & v2); //ʸ
VEC_TEMPLATE ORBITDYN_API Vector operator-(const Vector& v1,const Vector& v2); //ʸ
VEC_TEMPLATE ORBITDYN_API Vector operator-=(Vector & v1,const Vector & v2); //ʸ
VEC_TEMPLATE ORBITDYN_API Vector operator-(const Vector& v);//ʸ
MAT_TEMPLATE ORBITDYN_API Vector operator*(const matrixT& m,const Vector& v);//3*3ʸ
MAT_TEMPLATE ORBITDYN_API Vector operator+(const matrixT& m,const Vector& v);//3*1ʸ
MAT_TEMPLATE ORBITDYN_API Vector operator+(const Vector& v,const matrixT& m);//ʸ3*1
MAT_TEMPLATE ORBITDYN_API Vector operator-(const matrixT& m,const Vector& v);//3*1ʸ
MAT_TEMPLATE ORBITDYN_API Vector operator-(const Vector& v,const matrixT& m);//ʸ3*1

#if defined _MSC_VER && _MSC_VER>1300
// in MSVC7.1 this can be compilered, but not in MSVC6.0
ORBITDYN_API Vector operator*(const Vector& v,const double num); //ʸ
ORBITDYN_API Vector operator*(const double num,const Vector& v); //ʸ

#else
// in MSVC6.0 

ORBITDYN_API Vector operator*(const Vector& v,const double num); //ʸ
ORBITDYN_API Vector operator*(const double num,const Vector& v); //ʸ

#endif

ORBITDYN_API double operator*(const Vector& v1,const Vector& v2);//
ORBITDYN_API Vector operator^(const Vector& v1,const Vector& v2);//
ORBITDYN_API Vector operator/(const Vector& v,const double num); //ʸ
ORBITDYN_API std::ostream & operator<< (std::ostream & ostrm, const Vector& v);
ORBITDYN_API std::istream & operator>> (std::istream & istrm,const Vector& v);


#endif //__VECTOR_H
