/*
	Filename: 	CSatellite.cpp
	Created:	2003/10/29
	Author:		HanDle
*/

#include "Satellite.h"
#include "Utility.h"

unsigned int CSatelliteBase::Counter=0;

CSatelliteBase::CSatelliteBase(void)
	:ElapsedTime(0.0),Mass0(1000.0),BurnedFuel(0.0)
	,Cd(2.0),AirDragArea(20.0),Cr(1.0),LightPressArea(20.0)
	,ThrustIsOn(false),ThrustDirection(1,0,0)
	,LL(6),Perturbation(ODP_EARTH_ALL),FirstTimeOutput(true),Save(false)
{
	Counter++;
	Name = char(96+Counter);
	Engine = new CEngine_const(270,20);
}

CSatelliteBase::CSatelliteBase(std::string name)
	:ElapsedTime(0.0),Name(name),Mass0(1000.0),BurnedFuel(0.0)
	,Cd(2.0),AirDragArea(20.0),Cr(1.0),LightPressArea(20.0)
	,ThrustIsOn(false),ThrustDirection(1,0,0)
	,LL(6),Perturbation(ODP_EARTH_ALL),FirstTimeOutput(true),Save(false)
{
	Counter++;
	Engine = new CEngine_const(270,20);
}

CSatelliteBase::~CSatelliteBase()
{
	delete Engine;
	if(FileInstElem.is_open()) FileInstElem.close();
	Counter--;
}
/*! ʼ
ElapsedTime
BurnedFuel
FileInstElemѾ򿪣򼴽رաдµʱὫպд
FirstTimeOutputΪtrue
#Initialize()Ӧõô˺Ա֤ʼȫ#
*/
void CSatelliteBase::Initialize()
{
	ElapsedTime = 0;
	BurnedFuel = 0;
	ThrustIsOn = false;
	if(FileInstElem.is_open())
		FileInstElem.close();
	FirstTimeOutput = true;
	RefreshStatus();
}

/*! ʼ
ElapsedTime
BurnedFuel
FileInstElemѾ򿪣򼴽رաдµʱὫպд
FirstTimeOutputΪtrue
#Initialize()Ӧõô˺Ա֤ʼȫ#
*/
void CSatelliteBase::Initialize(const CDateTime& t,const Vector& r,const Vector& v)
{
	Epoch = t;
	Position = r;
	Velocity = v;
	ElapsedTime = 0;
	BurnedFuel = 0;
	ThrustIsOn = false;
	if(FileInstElem.is_open())
		FileInstElem.close();
	FirstTimeOutput = true;
	RefreshStatus();
}

/*!ļȡʼ
*/
void CSatelliteBase::Initialize( const char* filename )
{
	char buf[1024],strbuf[100],equator;
	Kepler kp;
	fstream rorbit(filename,ios::in);
	if(!rorbit.is_open()){
		throw SatelliteException("CSatelliteBase::Initialize( const char* filename ) : can't find file");
	}

	while(!rorbit.eof())
	{
		rorbit.getline(buf,1024);
		sscanf_s(buf,"%s",strbuf);
		_strupr_s(strbuf); // תΪдĸ
		std::string str(strbuf);
		if(str == "CENTER_NAME") {
			sscanf_s(buf,"%s %s %s",strbuf,strbuf,strbuf);
			if( !strcmp(strbuf,"EARTH") ) center = EARTH;
			if( !strcmp(strbuf,"MOON")  ) center = MOON;
		}
		else if(str == "EPOCH")		{
			sscanf_s(buf,"%s %s %s",strbuf,strbuf,strbuf);
			Epoch = StrToDateTime(strbuf);
		}
		else if(str == "SEMI_MAJOR_AXIS")	{
			sscanf_s(buf,"%s %c %lf",strbuf,&equator,&kp.a);  }
		else if(str == "ECCENTRICITY")		{
			sscanf_s(buf,"%s %c %lf",strbuf,&equator,&kp.e);  }
		else if(str == "INCLINATION")		{
			sscanf_s(buf,"%s %c %lf",strbuf,&equator,&kp.i);    kp.i *= RAD; }
		else if(str == "RA_OF_ASC_NODE")	{
			sscanf_s(buf,"%s %c %lf",strbuf,&equator,&kp.o);	kp.o *= RAD; }
		else if(str == "ARG_OF_PERICENTER")	{
			sscanf_s(buf,"%s %c %lf",strbuf,&equator,&kp.w);	kp.w *= RAD; }
		else if(str == "MEAN_ANOMALY")		{
			sscanf_s(buf,"%s %c %lf",strbuf,&equator,&kp.M);	kp.M *= RAD; }
		else if(str == "MASS")		{
			sscanf_s(buf,"%s %c %lf",strbuf,&equator,&Mass0);  }
	}
	rorbit.close();

	Kepler_Cartesian(kp,Position,Velocity);
	ElapsedTime = 0;
	BurnedFuel = 0;
	ThrustIsOn = false;
	if(FileInstElem.is_open())
		FileInstElem.close();
	FirstTimeOutput = true;
	RefreshStatus();
}

/*! Ҫǵ㶯
\param l Ľ
\param PTB Ҫǵ㶯ı־ĬΪODP_LEO
*/
void CSatelliteBase::SetForce(int l,unsigned int PTB)
{
	LL = l;
	Perturbation = PTB;
}

/*! ÷ıȳ
\param I ȳ
\param f 
*/
void CSatelliteBase::SetEngine(double I,double f)
{
	Engine->Isp = I;
	Engine->F = f;
}

/*! ǷԶ
SetAutoSave()  SetAutoSave(true) Զ\n
SetAutoSave(false)   Զ
*/
void CSatelliteBase::SetAutoSave(bool save /* =true */)
{
	Save = save;
}
/*!  Զ
ԶĸʽΪ\n
ʱt       볤a    ƫe    i    ྭOmega     صw     ƽM      u
*/
void CSatellite::SaveElem()
{
#define SPACE  "     "
	if(!FileInstElem.is_open())
	{
		std::string finst = Name + ".dat";
		const char * fi = finst.c_str();
		FileInstElem.open(fi,ios::out|ios::trunc);
	}
	FileInstElem.setf(ios::fixed);
	FileInstElem.precision(3);
	FileInstElem.width(11);
	FileInstElem<<ElapsedTime<<SPACE;
	FileInstElem.precision(5);
	FileInstElem.width(11);
	FileInstElem<<a<<SPACE;
	FileInstElem.width(11);
	FileInstElem.precision(8);
	FileInstElem<<e<<SPACE;
	FileInstElem.width(9);
	FileInstElem<<i*DEG<<SPACE;
	FileInstElem.width(9);
	FileInstElem<<Omega*DEG<<SPACE;
	FileInstElem.width(9);
	FileInstElem<<w*DEG<<SPACE;
	FileInstElem.width(9);
	FileInstElem<<M*DEG<<SPACE;
	FileInstElem.width(9);
	FileInstElem<<u*DEG<<endl;
#undef SPACE
}

/*! һƣֹͣ
	һ˺Ĺ˳Ϊ
- 1. ״̬x
- 2. RKF78һ״̬xh
- 3. ֹͣȷǷxhǵ״̬\n
     ֹͣСٴεRKF78,ֱﵽֹͣ
- 4. Ϊ䲽֣RKF78صĽض޸Ĳ

\nҪش˺밴ɡ
*/
int CSatelliteBase::OneStep(double StepSize,double MaxCutError /* = 1 */,CStopCondition* StopCon/* = NULL */)
{
	double x[7],xh[7],ee;
	Position.ToArray(x);
	Velocity.ToArray(x+3);
	x[6] = Mass();
	ee = RKF78(*this,StepSize,x,xh,7);

	if(StopCon == NULL)
	{
		ElapsedTime += StepSize;
		Position.SetValue(xh);
		Velocity.SetValue(xh+3);
		BurnedFuel += Mass() - xh[6];
		RefreshStatus();		
	}
	else
	{
		// ֹͣж
	}

	double y[7];
	DynFunction(ElapsedTime,xh,y,7);
	acc.SetValue(y[3],y[4],y[5]);

	return 1;
}

/*! dtʱ
\param h Ƶļ㲽()
\param dt Ƶʱ()
йֱӵһ˺ɣ60Ĳһã\n
Satellite.Propagate(60,86400); \n
һҪ鿴һĹݣҪԶ: \n
Satellite.SetAutoSave(true); \n
ɺֱӶȡԶļ\n\n
ѹƷŵѭУʹ:\n
Satellite.Propagate(Step,Step); \n
StepΪѧ/沽/
*/
void CSatelliteBase::Propagate(double h,const double dt)
{	
	double te = ElapsedTime + dt;
	bool calculate = true;
	if(FirstTimeOutput){
		if(Save) SaveElem();
		FirstTimeOutput = false;
	}

	while(calculate)
	{
		if(te-ElapsedTime<h)
		{
			calculate = false;
			h = te - ElapsedTime;
			if(h<1e-9) return;
		}
		OneStep(h);
		if(Save) SaveElem();
	}
}
/*!

*/
void CSatelliteBase::PropagateBackward(double h,const double dt)
{
	double te = ElapsedTime + dt;
	bool calculate = true;
	if(FirstTimeOutput){
		if(Save) SaveElem();
		FirstTimeOutput = false;
	}

	while(calculate)
	{
		if(te-ElapsedTime>h)
		{
			calculate = false;
			h = te - ElapsedTime;
			if(fabs(h)<1e-9) return;
		}
		OneStep(h);
		if(Save) SaveElem();
	}
}

/*! Ƶĵ
*/
void CSatelliteBase::Propagate2Perigee()
{
	double Step = 10;
	double pM;
	bool first = true , endwile = false;
	while( true )
	{
		Propagate(Step,Step);

		if(endwile)
			break;

		if(fabs(PI2-M)<15.0*RAD)
		{
			if(first)
			{
				first = false;
			}
			else
			{
				double n = (M - pM)/Step;
				if( M+n*Step>PI2 )
				{
					endwile = true;
					Step = fabs( (PI2-M) / n); // һı䲽ʹMպΪ0ĵ
				}
			}
			pM = M;
		}
	}
}

/*! ƵԶĵ
*/
void CSatelliteBase::Propagate2Apogee()
{
	double Step = 10;
	double pM;
	bool first = true , endwile = false;
	while( true )
	{
		Propagate(Step,Step);

		if(endwile)
			break;

		if( PI-M>0.0 && PI-M<10.0*RAD)
		{
			if(first)
			{
				first = false;
			}
			else
			{
				double n = (M - pM)/Step;
				if( M+n*Step>PI )
				{
					endwile = true;
					Step = fabs( (PI-M) / n); // һı䲽ʹMպΪPIԶĵ
				}
			}
			pM = M;
		}
	}
}

/*! Ƶ
*/
void CSatelliteBase::Propagate2AscendingNode()
{
	double Step = 10;
	double pu;
	bool first = true , endwile = false;
	while( true )
	{
		Propagate(Step,Step);

		if(endwile)
			break;

		if(fabs(PI2-u)<15.0*RAD)
		{
			if(first)
			{
				first = false;
			}
			else
			{
				double n = (u - pu)/Step;
				if( u+n*Step>PI2 )
				{
					endwile = true;
					Step = fabs( (PI2-u) / n); // һı䲽ʹuպΪ0
				}
			}
			pu = u;
		}
	}
}

/*! Ƶ
*/
void CSatelliteBase::Propagate2DescendingNode()
{
	double Step = 10;
	double pu;
	bool first = true , endwile = false;
	while( true )
	{
		Propagate(Step,Step);

		if(endwile)
			break;

		if( PI-u>0.0 && PI-u<10.0*RAD)
		{
			if(first)
			{
				first = false;
			}
			else
			{
				double n = (u - pu)/Step;
				if( u+n*Step>PI )
				{
					endwile = true;
					Step = fabs( (PI-u) / n); // һı䲽ʹuպΪPIｵ
				}
			}
			pu = u;
		}
	}
}

/*!ƵһԪ*/
void CSatelliteBase::Propagate2Epoch(const CDateTime epoch)
{
	if(epoch < CurrentEpoch() )
		PropagateBackward(-100,epoch-CurrentEpoch());
	else
		Propagate(100,epoch-CurrentEpoch());
}

/*! ˲ʱ
Input:  Dv:ٶ(km/s)  d:ٶϵ

Output:   None
Return:     None
*/
void CSatelliteBase::ImpluseManeuver(Vector Dv,Coordination d /*= VVLH*/ )
{
	switch(d) {
	case VVLH:
		Velocity += VVLH_ECI(Dv,*this);
		break;
	case ECI:
		Velocity += Dv;
		break;
	default:
		throw SatelliteException("ImpluseManeuver(): Can't find Operation in this Direction,Maneuver is ignored!");
	}
	RefreshStatus();
}

ORBITDYN_API ostream & operator<<(ostream & os, const CSatelliteBase& sat)
{
	os<<sat.a<<"\t"<<sat.e<<"\t"<<sat.i*DEG<<"\t"
	  <<sat.Omega*DEG<<"\t"<<sat.w*DEG<<"\t"<<sat.M*DEG;
	return os;
}

/*! ʼ
\param elem a,e,i,O,w,M
\param flag ƽ˲ʱı־
flag = 'i' ʾ˲ʱ\n
flag = 'm' ʾƽ\n
*/
void CSatellite::Initialize(const CDateTime& t,const Kepler elem,const char flag/*='i'*/)
{
	Epoch = t;
	if(flag=='i')
	{
		Status0 = elem;
	}
	else if(flag=='m')
	{
		Status0 = Instant(elem);
	}
	a = Status0.a;
	e = Status0.e;
	i = Status0.i;
	Omega = Status0.o;
	w = Status0.w;
	M = Status0.M;
	E = Status0.E();

	f = Status0.f();
	u = fmod(f + w + PI2,PI2);
	n = Status0.n();

	Kepler_Cartesian(Status0,Position,Velocity);
	
	CSatelliteBase::Initialize();
	
}

/*! ɳʼı
*/
//void CSatellite::ReportInitial()
//{
//	std::string finit = Name + "Initial.dat";
//	const char * fi = finit.c_str();
//	fstream file(fi,ios::out);
	//fprintf(InitCon,"\t\t%s\n%s\n\n%s\n\n%s%s\n","Propagator Initial Conditions",
	//	"------------------------------------------------------------------------------",
	//	"Propagator Name = EarthFull","Gravity File Name = ",GrvFile);
	//fprintf(InitCon,"%s%d%s%d\n","Gravity Degree = ",LL," X ",LL);
	//fprintf(InitCon,"%s%s\n\n","Include Tesseral Terms?  ",P_Tesseral?"Yes":"No");
	//fprintf(InitCon,"%s%s\n","Include Lunar Gravity?  ",P_Lunar?"Yes":"No");
	//fprintf(InitCon,"%s%s\n\n","Include Solar Gravity?  ",P_Solar?"Yes":"No");
	//fprintf(InitCon,"%s%s\n\n","Include Post-Newton?  ",P_PostNewton?"Yes":"No");
	//fprintf(InitCon,"%s%s\n","Include Air Drag?  ",P_Air?"Yes":"No");
	////if(P_Air) fprintf(InitCon,"%s%s\n","Atmosphere Density File = ",AtmosFile);
	//if(P_Air) fprintf(InitCon,"%s%lf\n","DragArea = ",AirDragArea);
	//fprintf(InitCon,"\n%s%s\n","Include Light Pressure?  ",P_Light?"Yes":"No");
	//if(P_Light) fprintf(InitCon,"%s%lf\n\n","LightArea = ",LightPressArea);
	//fprintf(InitCon,"%s%lf\n\n","CSatellite Mass = ",DryMass);
	////fprintf(InitCon,"%s%d%s%d%s%d%s%d%s%d%s%lf\n","Orbit Epoch : ",StartEpoch.year,"-",StartEpoch.month,"-",StartEpoch.day,
	////	" ",StartEpoch.hour,":",StartEpoch.minute,":",StartEpoch.second);
	//fprintf(InitCon,"%s\n\t%s%lf\n","Initial Orbit Elements:","Semi-major Axis = ",a*Re);
	//fprintf(InitCon,"\t%s%lf\n","Eccentricity = ",e);
	//fprintf(InitCon,"\t%s%lf\n","Inclination = ",i/RAD);
	//fprintf(InitCon,"\t%s%lf\n","RAAN = ",Omega/RAD);
	//fprintf(InitCon,"\t%s%lf\n","Angle of Periapsis = ",w/RAD);
	//fprintf(InitCon,"\t%s%lf\n","Mean Anomaly = ",M/RAD);
	//fprintf(InitCon,"\n%s\n","Integrator = RKF7(8)");
	////fprintf(InitCon,"%s%lf","Initial Integrator Step Size = ",h);
	////if(int(dd)==1) fprintf(InitCon,"\t%s","Fixed");
	////	else fprintf(InitCon,"\n\t%s%lf","Relative Error Tolerance = ",dd);
	//fclose(InitCon);
//	file.close();
//}

void CSatellite::GetECF(Vector& ECFr,Vector& ECFv)
{
	Matrix hg = ECI_ECF(Epoch+ElapsedTime);
	Vector w(0,0,We);
	w = PolarMotion(Epoch+ElapsedTime)*w;

	ECFr = hg*Position;
	ECFv = hg*Velocity - Cross(w,ECFr);
}

/*!
   ÿ״̬PositionVelocity仯ʱô˺ǹ
*/
void CSatellite::RefreshStatus()
{
	Kepler newkepler = Cartesian_Kepler(Position,Velocity);
	a = newkepler.a;
	e = newkepler.e;
	i = newkepler.i;
	Omega = newkepler.o;
	w = newkepler.w;
	M = newkepler.M;

	E = newkepler.E();
	f = newkepler.f();
	u = fmod(f + w + PI2,PI2);
	n = newkepler.n();
	r = Position.Module();
	v = Velocity.Module();
	w0 = sqrt(GE/a/a/a)*(1+2*e*cos(f)+e*e*cos(f)*cos(f))/sqrt(pow(1-e*e,3));
	Coi = GetCoi(newkepler);
}
/*! ƽ
*/
Kepler CSatellite::MedianElement() const
{
	Kepler Inst(a,e,i,Omega,w,M);
	return Mean(Inst);
}
/*! ѧ
   Input:  t:ʱ   x[6]:״̬
   Output:    y[6]:״̬΢
*/
void CSatellite::DynFunction(const double t,const double* x,double* y,int dim) const
{
	Vector F,rs,FG,ECFr,F12,r,rd;

	r.x = x[0];
	r.y = x[1];
	r.z = x[2];
	rd.x = x[3];
	rd.y = x[4];
	rd.z = x[5];

	CDateTime tt = Epoch + t;

	// Zonal Harmonic Terms and Tesseral Harmonic Terms
	Matrix HG = ECI_ECF(tt);
	ECFr =  HG * r;  //J2000.0==>ECF  //˴ΪǳʼԪֵ

	if(Perturbation & ODP_EARTH_ZONAL)
	{
		FG = PZonalHarmonic(LL,ECFr);
		if(Perturbation & ODP_EARTH_TESSERAL)
		{
			FG += PTesseralHarmonic(LL,ECFr);
		}
		F12 = ~HG * FG;  //ECF==>J2000.0   //˴ΪǳʼԪֵ
		F = F + F12;
	}


	// Lunar Gravitation
	if(Perturbation & ODP_LUNAR_CENT)
	{
		F += PThirdBodyGrv( r , Moon::Instance()->GetECI(tt) , GM );
	}

	// Solar Gravitation
	if( (Perturbation & ODP_SOLAR_CENT)  || (Perturbation & ODP_SOLAR_PRESSURE) || (Perturbation & ODP_AIR_DRAG) )
	{
		rs = Sun::Instance()->GetECI(tt);   // rs:solar coordinate in J2000.0
	}

	// solar gravity
	if( Perturbation & ODP_SOLAR_CENT )
	{
		F += PThirdBodyGrv(r,rs,GS);
	}

	// Sunlight Pressure
	if( Perturbation & ODP_SOLAR_PRESSURE )
	{
		F += PLightPressure(Cr*LightPressArea/(Mass0-BurnedFuel),r,rs);
	}

	// Atmospheric Drag
	if( Perturbation & ODP_AIR_DRAG )
	{
		F += PAirDrag(tt,Cd*AirDragArea/(Mass0-BurnedFuel),ECFr,r,rd);
	}

 	// Post-Newtonian Effect
	if( Perturbation & ODP_POSTNEWTON )
	{
		F += PPostNewton(r,rd);
	}

	//if(Perturbation & ODP_LUNAR_NON_SPHERE)
	//{   // һڵתƹӰ֮
	//	F += PLunar_NonSphere(TDT,ECI_MCI(r,TDT));
	//}

	// Thrust
	if(ThrustIsOn)
	{
		F += ~(qbi.C()) * ThrustDirection * Engine->F/(Mass0-BurnedFuel)/1000.0;
	}

	// Perturbation Functions
	double R = r.Module();//,v = rd.Module();
	double miurrr = GE/R/R/R;
	y[0] = x[3];
	y[1] = x[4];
	y[2] = x[5];
	y[3] = - miurrr * r.x + F.x;
	y[4] = - miurrr * r.y + F.y;
	y[5] = - miurrr * r.z + F.z;
	if( ThrustIsOn )
		y[6] = Engine->dm();
	else
		y[6] = 0;
}

/*!  ǵĵعϵ
*/
const Vector CSatellite::ECFPos()
{
	return ECI_ECF(Epoch + ElapsedTime) * Position;
}
/*! ǵľγ()
*/
CSpherical CSatellite::GetLLA()
{
	Vector ECF = ECFPos(); // عϵλ
	return ECF_LLA(ECF);
}
/*! ɸµ㾭γ
*/
CSpherical CSatellite::SubSatPoint(CSpherical & Geodetic,CSpherical & SSP)
{
	double r = a*(1 - e * cos(E));
	// 
	double lmdG = GetSG(Epoch+ElapsedTime);
	Geodetic.Longitude = fmod(atan2(Position.y,Position.x)-lmdG+4*PI,PI2);
	if(Geodetic.Longitude>PI) Geodetic.Longitude -= PI2;
	// γ
	double zsr = Position.z/r;
	double Lan,BB,Rs0,hh;
	if( zsr<-1 ) Geodetic.Latitude = -90;
	else if( zsr>1 ) Geodetic.Latitude = 90;
	else Geodetic.Latitude = asin(zsr);
	Lan = Geodetic.Latitude;

	// µ㾭
	SSP.Longitude = Geodetic.Longitude;
	// µγ
	double Lane = Geodetic.Latitude;
	double Ob1 = 1-Oblate;
	for(int II=1;II<=4;II++)
	{
		double sinLane = sin(Lane);
		double cosLane = cos(Lane);
		BB = atan(tan(Lane)/Ob1/Ob1);
		Rs0 = Re*Ob1/sqrt(sinLane*sinLane+Ob1*Ob1*cosLane*cosLane);
		hh = sqrt(r*r-(Rs0*Rs0*sin(BB-Lane)*sin(BB-Lane)))-Rs0*cos(BB-Lane);
		Lane = Lan-asin(hh*sin(BB-Lane)/r);
	}
	Geodetic.Altitude = hh;
	SSP.Latitude = Lane;
	SSP.Altitude = 0;
	return SSP;
}

//double CSatellite::Propagate2u(double &t,double h,CSatellite &s,double ut,const double dd,FILE* out)
//{//failure
//	double x1[6],x[6],tmp[6],tt;
//	s.mean();
//	x[0]=s.M_a;x[1]=s.M_i;x[2]=s.M_omiga;x[3]=s.M_ksi;x[4]=s.M_eta;x[5]=s.M_lamda;
//	x1[0]=s.a;x1[1]=s.i;x1[2]=s.Omega;x1[3]=s.ksi;x1[4]=s.eta;x1[5]=s.lamda;
//	tt=t;
//	for(int i=0;i<6;i++) tmp[i]=x1[i];
////----lamda1--------------------------------------------------------------------------
//	double n=sqrt(1/x[0]/x[0]/x[0]); //(unit:rad/s)
//	double sini2=sin(x[1])*sin(x[1]);
//	double sini4=sini2*sini2;
//	double e=sqrt(x[3]*x[3]+x[4]*x[4]);
//	double ee=sqrt(1-x[3]*x[3]-x[4]*x[4]);
//	double lamda1=A2/x[0]/ee*n*((2-2.5*sini2)+(1-1.5*sini2)*ee);
////----end lamda1---------------------------------------------------------------------
//	//double sinu=sin(ut);
//	//double cosu=cos(ut);
//	s.other('m');
//	double u1=s.M_u;
//	double du=ut-u1;
//	du=du>0?du:(du+PI2);
////Сƫʹdf+dwԼdM+dwԼdlamda
//	double dt1=du/(n+lamda1);   //(unit:minute)
//	double dt=dt1;
//	double dt2=dt1;
//	while(fabs(dt2)>1/60.0/TU)
//	{
//		Propagate2T(tt,h,x1,dt,dd,NULL);
//		//double Ew1=kepler2(x1[3],x1[4],x1[5]);
//		//double w1=fmod(atan2(x1[4],x1[3])+PI2,PI2);
//		//double E1=fmod(Ew1-w1+4*PI,PI2);
//		//double e1=sqrt(x1[3]*x1[3]+x1[4]*x1[4]);
//		//double r1=x1[0]*(1-e1*cos(E1));
//		//double ee1=sqrt(1-e1*e1);
//		CSatellite ttt;
//		ttt.a=x1[0];ttt.i=x1[1];ttt.Omega=x1[2];ttt.ksi=x1[3];ttt.eta=x1[4];ttt.lamda=x1[5];
//		ttt.elem_to_xyz();
//		dt2=-ttt.z/ttt.dz;
//		//dt2=-(r1*sqrt(x1[0]))*((cos(E1)-e1)*sin(w1)-ee1*cos(w1))/(-sin(E1)*sin(w1)+ee1*cos(E1)*cos(w1));
//		dt=dt+dt2;  //refresh dt
//		for(int ii=0;ii<6;ii++) x1[ii]=tmp[ii];  //return to start state
//		tt=t;                                    //return to start state
//	}
//	Propagate2T(t,h,x,dt,dd,out);
//	s.a=x[0];s.i=x[1];s.Omega=x[2];s.ksi=x[3];s.eta=x[4];s.lamda=x[5];
//	return dt;
//}

//double CSatellite::Propagate2M(double &t,double h,CSatellite &s,const double Mt,const double dd,FILE* out)	//ƵƽM		|
//{
//	double n=sqrt(1/s.a/s.a/s.a);
//	s.other('i');
//	double dM=s.M-Mt;
//	dM=dM>0?dM:(dM+PI2);
//	double dt=dM/n;
//
//	double tt=t;
//	double lx[6];
//	lx[0]=s.a;lx[1]=s.i;lx[2]=s.Omega;lx[3]=s.ksi;lx[4]=s.eta;lx[5]=s.lamda;
//	while(true)
//	{
//		Propagate2T(tt,h,s,dt,dd,out);
//	}
//}

//! ʼú
void CRapidSatellite::Initialize(const CDateTime& t,const Kepler elem,const char flag)
{
	ElapsedTimems = 0;
	ElapsedTime = 0;
	m_lastpert = 0;
	m_lasttime = 0;
	CSatellite::Initialize(t,elem,flag);
	// ʼʱ̵ת㶯ٶ
	RefreshMatrix(Epoch);
	RefreshPert(Epoch);
}

void CRapidSatellite::RefreshMatrix( const CDateTime tt )
{
	PR = Precession(tt);
	NU = Nutation(tt);
	PW = PolarMotion(tt);
}

void CRapidSatellite::RefreshPert( const CDateTime tt )
{
	HG = PW*EarthRotation(tt)*NU*PR;
	Vector ECFr =  HG * Position;  //J2000.0==>ECF  //˴ΪǳʼԪֵ
	if(Perturbation & ODP_EARTH_ZONAL)
	{
		F_zonal = PZonalHarmonic(LL,ECFr);
		F_zonal = ~HG*F_zonal;
		if(Perturbation & ODP_EARTH_TESSERAL)
		{
			F_tesseral = PTesseralHarmonic(LL,ECFr);
			F_tesseral = ~HG*F_tesseral;
		}
	}
	rm = Moon::Instance()->GetECI(tt);
	rs = Sun::Instance()->GetECI(tt);
	// Lunar Gravitation
	if(Perturbation & ODP_LUNAR_CENT)
	{
		F_moon = PThirdBodyGrv(Position,rm,GM);
	}

	// solar gravity
	if( Perturbation & ODP_SOLAR_CENT )
	{
		F_sun = PThirdBodyGrv(Position,rs,GS);
	}

	// Sunlight Pressure
	if( Perturbation & ODP_SOLAR_PRESSURE )
	{
		F_light = PLightPressure(Cr*LightPressArea/(Mass0-BurnedFuel),Position,rs);
	}

	// Atmospheric Drag
	if( Perturbation & ODP_AIR_DRAG )
	{
		F_air = PAirDrag(tt,Cd*AirDragArea/(Mass0-BurnedFuel),ECFr,Position,Velocity);
	}

	// Post-Newtonian Effect
	if( Perturbation & ODP_POSTNEWTON )
	{
		F_ppn = PPostNewton(Position,Velocity);
	}
}
// hΪƲλΪms
void CRapidSatellite::Propagate(int h)
{	
	if(FirstTimeOutput){
		if(Save) SaveElem();
		FirstTimeOutput = false;
	}

	OneStep(h);
	if(Save) SaveElem();
}

// StepSizeλΪms
int CRapidSatellite::OneStep(int StepSize,double MaxCutError /* = 1 */,CStopCondition* StopCon/* = NULL */)
{
	double x[7],xh[7],ee;
	Position.ToArray(x);
	Velocity.ToArray(x+3);
	x[6] = Mass();

	CDateTime tt = Epoch + ElapsedTimems/1000.0;

	if( _abs64(ElapsedTimems - m_lasttime) >= 1800000 )
	{
		m_lasttime = ElapsedTimems;
		RefreshMatrix(tt);
		//HG = ECI_ECF(tt);
	}
	// ³ʼԺElapsedTimeΪ0m_lastpertԭֵ㶯ûи
	//if(ElapsedTime>=m_lastpert+0.05-DBL_EPSILON
	// ڸΪһ䣬֤ElapsedTimem_lastpertĲһʱԻ㶯
	if( _abs64(ElapsedTimems - m_lastpert) >= PeriodPert)
	{
		m_lastpert = ElapsedTimems;

		RefreshPert(tt);
	}

	// Thrust
	if(ThrustIsOn)
	{
		F_Thrust = ~(qbi.C()) * ThrustDirection * Engine->F/(Mass0-BurnedFuel)/1000.0;
	}

	//ee = RKF78(*this,StepSize,x,xh,7);
	ee = Gill4(*this,StepSize/1000.0,x,xh,7);

	if(StopCon == NULL)
	{
		ElapsedTimems += StepSize;
		ElapsedTime = ElapsedTimems/1000.0;
		Position.SetValue(xh);
		Velocity.SetValue(xh+3);
		BurnedFuel += Mass() - xh[6];
		RefreshStatus();
	}
	else
	{
		// ֹͣж
	}

	double y[7];
	DynFunction(ElapsedTime,xh,y,7); //ElapsedTimeʵû
	acc.SetValue(y[3],y[4],y[5]);

	return 1;
}

/*! ѧ
Input:  t:ʱ   x[6]:״̬
Output:    y[6]:״̬΢
*/
void CRapidSatellite::DynFunction(const double t,const double* x,double* y,int dim) const
{
	Vector F;

	if(Perturbation & ODP_EARTH_ZONAL)
	{
		F += F_zonal;
	}

	if(Perturbation & ODP_EARTH_TESSERAL)
	{
		F += F_tesseral;
	}

	// Lunar Gravitation
	if(Perturbation & ODP_LUNAR_CENT)
	{
		F += F_moon;
	}

	// solar gravity
	if( Perturbation & ODP_SOLAR_CENT )
	{
		F += F_sun;
	}

	// Sunlight Pressure
	if( Perturbation & ODP_SOLAR_PRESSURE )
	{
		F += F_light;
	}

	// Atmospheric Drag
	if( Perturbation & ODP_AIR_DRAG )
	{
		F += F_air;
	}

	// Post-Newtonian Effect
	if( Perturbation & ODP_POSTNEWTON )
	{
		F += F_ppn;
	}

	// Thrust
	if(ThrustIsOn)
	{
		F += F_Thrust;
	}

	// Perturbation Functions
	double R = sqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
	double miurrr = GE/R/R/R;
	y[0] = x[3];
	y[1] = x[4];
	y[2] = x[5];
	y[3] = - miurrr * x[0] + F.x;
	y[4] = - miurrr * x[1] + F.y;
	y[5] = - miurrr * x[2] + F.z;
	if( ThrustIsOn )
		y[6] = Engine->dm();
	else
		y[6] = 0;
}

/*!
ÿ״̬PositionVelocity仯ʱô˺ǹ
*/
void CRapidSatellite::RefreshStatus()
{
	Kepler newkepler = Cartesian_Kepler(Position,Velocity);
	a = newkepler.a;
	e = newkepler.e;
	i = newkepler.i;
	Omega = newkepler.o;
	w = newkepler.w;
	M = newkepler.M;

	E = newkepler.E();
	f = newkepler.f();
	u = fmod(f + w + PI2,PI2);
	n = newkepler.n();
	r = Position.Module();
	v = Velocity.Module();
	w0 = sqrt(GE/a/a/a)*(1+2*e*cos(f)+e*e*cos(f)*cos(f))/sqrt(pow(1-e*e,3));
	Coi = GetCoi(newkepler);
}

/*!
ʹСƫʹΪֱRKF78汾ʹõķ
ֻеĵλΪkm,s,rad
*/
int CLEO::OneStep(double StepSize,double MaxCutError /* = 1.0 */,CStopCondition* SC /* = NULL */)
{
	double x[7],xh[7],ee;
	// ֱΪ a  i  Omega  ksi  eta  lamda
	x[0] = a;
	x[1] = i;
	x[2] = Omega;
	x[3] = e*cos(w);
	x[4] = e*sin(w);
	x[5] = fmod(w+M+PI2,PI2);
	if(ThrustIsOn)
	{
		x[6] = Mass();
		ee = RKF78(*this,StepSize,x,xh,7);
	}
	else
	{
		ee = RKF78(*this,StepSize,x,xh,6);
	}
	if(SC == NULL)
	{
		ElapsedTime += StepSize;
		xh[2] = fmod(xh[2]+PI2,PI2);
		xh[5] = fmod(xh[5]+PI2,PI2);
		a = xh[0];
		i = xh[1];
		Omega = xh[2];
		e = sqrt(xh[3]*xh[3]+xh[4]*xh[4]);
		w = fmod(atan2(xh[4],xh[3])+PI2,PI2);
		M = fmod(xh[5]-w+PI2,PI2);
		//... λáٶ
		RefreshStatus();
		BurnedFuel += Mass() - xh[6];
	}
	else
	{

	}
	//if(MaxCutError>0.5) return 0;
	//if(ee>MaxCutError)  return StepSize = StepSize/2;
	//if(ee<MaxCutError*0.01) return StepSize*2;
	//return StepSize;
	return 1;
}
/*! ˺ʵCSatelliteBase::RefreshKepler()
*/
void CLEO::RefreshStatus()
{
	Kepler k(a,e,i,Omega,w,M);
	Kepler_Cartesian(k,Position,Velocity);

	E = k.E();
	f = k.f();
	u = fmod(f + w + PI2,PI2);
	n = k.n();
	r = Position.Module();
	v = Velocity.Module();
}

/*!ѧ
t:ʱ   x[6]:״̬a  i  Omega  ksi  eta  lamda
y[6]:״̬΢
*/
void CLEO::DynFunction(const double t,const double* x,double* y,int dim) const
{
	Vector F;
	static Vector rs,FG,ECFr,F12,r,rd,RR,TT,P,Q,S;
	// u
	double uE=KeplerFunc2(x[3],x[4],x[5]);
	double SuE=sin(uE);
	double CuE=cos(uE);
	double e2=x[3]*x[3]+x[4]*x[4];
	double GE2=sqrt(1.0-e2);
	double WE2=1.0+GE2;
	double ZuE=x[3]*CuE+x[4]*SuE;
	double FuE=x[3]*SuE-x[4]*CuE;
	double r0=x[0]*(1.0-ZuE); //ʸ
	double Gp=sqrt(x[0]*(1.0-e2));
	double Sdu=x[0]/r0*FuE*(1.0-ZuE/WE2);
	double Cdu=sqrt(1.0-Sdu*Sdu);
	double du=atan2(Sdu,Cdu);
	double u=uE+du;
	// 굥λʸ
	double Su=sin(u);
	double Cu=cos(u);
	double Si=sin(x[1]);
	double Ci=cos(x[1]);
	double SQ=sin(x[2]);
	double CQ=cos(x[2]);
	P.SetValue(CQ,SQ,0);
	Q.SetValue(-Ci*SQ,Ci*CQ,Si);
	RR=Cu*P+Su*Q;
	TT=-Su*P+Cu*Q;

	r = r0 * RR;
	rd = sqrt(GE)*Gp / r0 * (FuE/GE2*RR + TT);

	CDateTime tt = Epoch + t/86400;
	Matrix HG = ECI_ECF(tt);

	// Zonal Harmonic Terms and Tesseral Harmonic Terms
	ECFr = HG * r;	//J2000.0==>ECF  //˴ΪǳʼԪֵ
	if(Perturbation & ODP_EARTH_TESSERAL) FG = PZonalHarmonic(LL,ECFr) + PTesseralHarmonic(LL,ECFr);
	else FG = PZonalHarmonic(LL,ECFr);
	F12 = ~HG * FG; //ECF==>J2000.0   //˴ΪǳʼԪֵ
	F = F + F12;

	// Lunar Gravitation
	if(Perturbation & ODP_LUNAR_CENT)
		F += PThirdBodyGrv( r , Moon::Instance()->GetECI(tt) , GM );

	// Solar Gravitation
	if( (Perturbation & ODP_SOLAR_CENT)  || (Perturbation & ODP_SOLAR_PRESSURE) || (Perturbation & ODP_AIR_DRAG) )
		rs = Sun::Instance()->GetECI(tt);   // rs:solar coodinate in J2000.0

	if( Perturbation & ODP_SOLAR_CENT )
		F += PThirdBodyGrv(r,rs,GS);

	// Sunlight Pressure
	if(Perturbation & ODP_SOLAR_PRESSURE)
		F += PLightPressure(Cr*LightPressArea/(Mass0-BurnedFuel),r,rs);

	// Atmospheric Drag
	if( Perturbation & ODP_AIR_DRAG )
	{
		F += PAirDrag(tt,Cd*AirDragArea/(Mass0-BurnedFuel),ECFr,r,rd);
	}

	// Post-Newtonian Effect
	if( Perturbation & ODP_POSTNEWTON )
		F += PPostNewton(r,rd);

	// Thrust
	if(ThrustIsOn)
		F += VVLH_ECI(ThrustDirection*Engine->F/(Mass0-BurnedFuel)/1000.0,*this);  //ϵеٶ

	// All Perturbation Accelerations
	S = F.RotationZ(x[2]).RotationX(x[1]).RotationZ(u);

	// Perturbation Functions
	double An = sqrt(GE/x[0]/x[0]/x[0]);
	double Ana = An * x[0];
	double Ana2 = Ana * x[0];
	double r0p = r0/(Gp*Gp);
	double CuSu = x[3]*Cu+x[4]*Su;
	double SuCu = x[3]*Su-x[4]*Cu;
	y[0] = 2.0/(An*GE2)*(S.x*SuCu+S.y/r0p);
	y[1] = r0*Cu*S.z/(Ana2*GE2);
	y[2] = r0*Su*S.z/(Ana2*GE2)/Si;
	y[3] = x[4]*Ci*y[2]+GE2/Ana*(S.x*Su+S.y/GE2*(CuE+GE2*Cu-x[3]*ZuE/WE2));
	y[4] = -x[3]*Ci*y[2]-GE2/Ana*(S.x*Cu-S.y/GE2*(SuE+GE2*Su-x[4]*ZuE/WE2));
	y[5] = An-Ci*y[2]-GE2/Ana*(2.0*S.x*GE2*r0p+(S.x*CuSu-S.y*(1.0+r0p)*SuCu)/WE2);
	if( ThrustIsOn )
		y[6] = Engine->dm();
	else
		y[6] = 0;
}

/*! ʼ
\param elem Ĺϵa,e,i,O,w,M
*/
void CMoonSat::Initialize(const CDateTime& t,const Kepler elem)
{
	Epoch = t;
	Status0 = elem;
	a = Status0.a;
	e = Status0.e;
	i = Status0.i;
	Omega = Status0.o;
	w = Status0.w;
	M = Status0.M;
	E = Status0.E();

	f = Status0.f();
	u = fmod(f + w + PI2,PI2);
	n = Status0.n();

	Kepler_Cartesian(Status0,Position,Velocity,GM);

	CSatelliteBase::Initialize();
}

/*! ʼ
\param elem Ĺϵa,e,i,O,w,M
*/
void CMoonSat::Initialize(const CDateTime& t,const Vector p,const Vector v)
{
	CSatelliteBase::Initialize(t,p,v);
	RefreshStatus();
}

/*!
ÿ״̬PositionVelocity仯ʱô˺ǹ
*/
void CMoonSat::RefreshStatus()
{
	Kepler newkepler = Cartesian_Kepler(Position,Velocity,GM);
	a = newkepler.a;
	e = newkepler.e;
	i = newkepler.i;
	Omega = newkepler.o;
	w = newkepler.w;
	M = newkepler.M;

	E = newkepler.E();
	f = newkepler.f();
	u = fmod(f + w + PI2,PI2);
	n = newkepler.n(GM);
	r = Position.Module();
	v = Velocity.Module();
}

/*! ¹ѧ
\param t ʱ
\param x ״̬
\param y ״̬΢
*/
void CMoonSat::DynFunction(const double t,const double* x,double* y,int dim) const
{
	Vector F;
	Vector rs,FG,ECFr,F12,r,rd;

	r.x = x[0];
	r.y = x[1];
	r.z = x[2];
	rd.x = x[3];
	rd.y = x[4];
	rd.z = x[5];

	CDateTime tt = Epoch + t;

	// 
	Matrix MIF = MCI_MCF(tt); //Ĺϵ¹ϵת
	Vector SatMCF = MIF*r;
	if(Perturbation & ODP_LUNAR_TESSERAL)
		FG = PZonalHarmonic(LL,SatMCF,Moon::Instance()) + PTesseralHarmonic(LL,SatMCF,Moon::Instance());
	else if(Perturbation & ODP_LUNAR_ZONAL)
		FG = PZonalHarmonic(LL,SatMCF,Moon::Instance());
	else
		FG = Vector(0,0,0);

	F += ~MIF*FG;

	Vector Moon = Moon::Instance()->GetECI(tt);
	// earth gravity
	if(Perturbation & ODP_EARTH_CENT) //ODP_LUNARʾǷǵ 
	{	
		F += PThirdBodyGrv(r, -Moon , GE); // MCIλüϵλõķ
	}

	// Solar Gravitation
	if( (Perturbation & ODP_SOLAR_CENT)  || (Perturbation & ODP_SOLAR_PRESSURE))
	{
		rs = Sun::Instance()->GetECI(tt);   // rs:solar coordinate in J2000.0
		rs -= Moon; // Ĺϵ̫
    }

	// solar gravity
	if( Perturbation & ODP_SOLAR_CENT )
	{
		F += PThirdBodyGrv(r,rs,GS);
	}

	// Sunlight Pressure
	if( Perturbation & ODP_SOLAR_PRESSURE )
	{		
		F += PLightPressure(Cr*LightPressArea/(Mass0-BurnedFuel),r,rs,Rm);
	}

	// Thrust
	if(ThrustIsOn)
	{
		F += ~(qbi.C()) * ThrustDirection * Engine->F/(Mass0-BurnedFuel)/1000.0;
	}

	// Perturbation Functions
	double R = r.Module();
	double miurrr = GM/R/R/R;
	y[0] = x[3];
	y[1] = x[4];
	y[2] = x[5];
	y[3] = - miurrr * r.x + F.x;
	y[4] = - miurrr * r.y + F.y;
	y[5] = - miurrr * r.z + F.z;
	if( ThrustIsOn )
		y[6] = Engine->dm();
	else
		y[6] = 0;
}

/*!  Զ
ԶĸʽΪ\n
ʱt       볤a    ƫe    i    ྭOmega     صw     ƽM      u
*/
void CMoonSat::SaveElem()
{
#define SPACE  "     "
	if(!FileInstElem.is_open())
	{
		std::string finst = Name + ".dat";
		const char * fi = finst.c_str();
		FileInstElem.open(fi,ios::out|ios::trunc);
	}
	FileInstElem.setf(ios::fixed);
	FileInstElem.precision(3);
	FileInstElem.width(11);
	FileInstElem<<ElapsedTime<<SPACE;
	FileInstElem.precision(5);
	FileInstElem.width(11);
	FileInstElem<<a<<SPACE;
	FileInstElem.width(11);
	FileInstElem.precision(9);
	FileInstElem<<e<<SPACE;
	FileInstElem.precision(8);
	FileInstElem.width(12);
	FileInstElem<<i*DEG<<SPACE;
	FileInstElem.width(12);
	FileInstElem<<Omega*DEG<<SPACE;
	FileInstElem.width(12);
	FileInstElem<<w*DEG<<SPACE;
	FileInstElem.width(12);
	FileInstElem<<M*DEG<<SPACE;
	FileInstElem.width(12);
	FileInstElem<<u*DEG<<endl;
#undef SPACE
}

